:- module influence_map.
:- interface.
:- import_module game.

	% "im" is short for "influence map".
	% An im records "influences" from objects of interest.
:- type im(T) == board(T).

	% A distance_im records distances from objects of interest.
	% The value at each square is the distance to the closest
	% object of interest.
	% `-1' is used to indicate an infinitely large distance
:- type distance == int.
:- type distance_im == im(distance).
:- func infinite_distance = distance. % returns -1

	% A closest_im(T) records which T is closest.
	% The value at each square is the identity of the closest
	% object of interest.
:- type closest_im(T) == im(T).

	% an enemy_distance_im stores the distance from each square to
	% the nearest enemy
:- type enemy_distance_im == distance_im. 
:- func compute_enemy_distance_im(game) = enemy_distance_im.

	% a base_im stores the distance from each square to the nearest base
:- type base_distance_im == distance_im.
:- func compute_base_distance_im(board) = base_distance_im.

	% A closest_base_im stores, for each square, the coordinates of
	% the close base to that square.  Squares which can't be reached
	% from a base are marked with `no_base_reachable'.
:- type nearest_base ---> no_base_reachable ; base_at(x, y).
:- type closest_base_im == closest_im(nearest_base).
:- func compute_closest_base_im(board) = closest_base_im.

	% compute_im(Board, InitFunc, IncrFunc, SentinelValue):
	%
	% compute_im builds an influence map from a board,
	% given an initialization function, an increment function,
	% and a sentinel value.
	% The initialization function should succeed for squares of
	% interest.  The increment function should propagate the
	% values to adjacent squares (failing for unreachable squares),
	% e.g. incrementing, decrementing, or copying, as required.
	% The value at each square in the final influence map will be
	% set to the value returned by the initialization function,
	% if any, or the value returned by applying the increment
	% function successively to neighbouring squares, starting with
	% a square.
	% The initialization function and the increment function
	% should never return the sentinel value.

:- type init_func(Sq,Val) == (func(x,y,Sq) = Val).
:- inst init_func == (func(in,in,in) = out is semidet).

:- type incr_func(Val) == (func(x,y,Val) = Val).
:- inst incr_func == (func(in,in,in) = out is semidet).

:- func compute_im(board(Sq), init_func(Sq,Val), incr_func(Val), Val) = im(Val).
:- mode compute_im(in, in(init_func), in(incr_func), in) = out is det.

:- func compute_distance_im(board, init_func(square_kind,distance)) =
	distance_im.
:- mode compute_distance_im(in, in(init_func)) = out is det.

%-----------------------------------------------------------------------------%

:- implementation.
:- import_module int, exception, bool, queue, list, map.

compute_enemy_distance_im(Game) =
	compute_distance_im(Game^board, enemy_init_func(Game)).

compute_base_distance_im(Board) =
	compute_distance_im(Board, base_distance_init_func).

compute_closest_base_im(Board) =
	compute_closest_im(Board, closest_base_init_func,
		no_base_reachable).

:- func enemy_init_func(game, x, y, square_kind) = distance is semidet.
enemy_init_func(Game, X, Y, _Square) = Distance :-
	% succeed if an enemy is present at this square
	list__member((X, Y), Game^enemy_robot_posns),
	Distance = 0.

:- func base_distance_init_func(x, y, square_kind) = distance is semidet.
base_distance_init_func(_X, _Y, Square) = Distance :-
	Square = home,
	Distance = 0.

:- func closest_base_init_func(x, y, square_kind) = nearest_base
	is semidet.
closest_base_init_func(X, Y, Square) = BasePosn :-
	Square = home,
	BasePosn = base_at(X, Y).

compute_distance_im(Board, InitFunc) =
	compute_im(Board, InitFunc, incr_distance(Board), infinite_distance).

:- func compute_closest_im(board, init_func(square_kind, T), T) = im(T).
:- mode compute_closest_im(in, in(init_func), in) = out is det.
compute_closest_im(Board, InitFunc, SentinalValue) =
	compute_im(Board, InitFunc, copy_closest(Board), SentinalValue).

:- func incr_distance(board, x, y, int) = int is semidet.
incr_distance(Board, X, Y, Dist) = Dist + 1 :-
	passable(square(Board, X, Y)) = yes.

:- func copy_closest(board, x, y, T) = T is semidet.
copy_closest(Board, X, Y, Target) = Target :-
	passable(square(Board, X, Y)) = yes.

:- func passable(square_kind) = bool.
passable(plain) = yes.
passable(home) = yes.
passable(lethal) = no.
passable(wall) = no.

infinite_distance = -1.

%-----------------------------------------------------------------------------%

compute_im(Board, InitFunc, IncrFunc, SentinelValue) = IM :-
	% Apply the InitFunc to each square of the board,
	% storing the resulting values in the initial influence map IM0.
	% If the InitFunc fails, store the SentinelValue.
	% Put all the squares for which the InitFunc succeeds
	% in the initial frontier.
	InitPred = (pred(X::in, Y::in, Sq::in, Influence::out,
			FrontierAcc0::in, FrontierAcc::out) is det :-
		( InitFunc(X, Y, Sq) = Influence0 ->
			Influence = Influence0,
			queue__put(FrontierAcc0, 
				frontier_elem(X, Y, Influence),
				FrontierAcc)
		;
			Influence = SentinelValue,
			FrontierAcc = FrontierAcc0
		)
	),
	queue__init(Frontier0),
	map_fold_board(InitPred, Board, IM0, Frontier0, Frontier1),
	% Now do a "flood fill" from the frontier.
	process_frontier(Frontier1, IncrFunc, SentinelValue, IM0, IM).

:- type frontier_elem(T) ---> frontier_elem(x, y, T).
:- type frontier(T) == queue(frontier_elem(T)).

:- mode process_frontier(in, in(incr_func), in,
		board_di, board_uo) is det.
process_frontier(Frontier0, IncrFunc, SentinelValue, IM0, IM) :-
	( queue__get(Frontier0, FrontierElem, Frontier1) ->
		FrontierElem = frontier_elem(x(X), y(Y), Influence),
		IM1 = IM0,
		process_neighbour(IncrFunc, Influence, SentinelValue,
			x(X+1),y(Y), Frontier1, Frontier2, IM1, IM2),
		process_neighbour(IncrFunc, Influence, SentinelValue,
			x(X-1),y(Y), Frontier2, Frontier3, IM2, IM3),
		process_neighbour(IncrFunc, Influence, SentinelValue,
			x(X),y(Y+1), Frontier3, Frontier4, IM3, IM4),
		process_neighbour(IncrFunc, Influence, SentinelValue,
			x(X),y(Y-1), Frontier4, Frontier5, IM4, IM5),
		process_frontier(Frontier5, IncrFunc, SentinelValue, IM5, IM)
	;
		% the frontier queue is empty, so we've finished
		IM = IM0
	).

:- pred process_neighbour(incr_func(T), T, T, x, y,
		frontier(T), frontier(T), im(T), im(T)).
:- mode process_neighbour(in(incr_func), in, in, in, in, in, out,
		board_di, board_uo) is det.

process_neighbour(IncrFunc, Influence0, SentinelValue, X, Y,
		Frontier0, Frontier, IM0, IM) :-
	OldValue = square(IM0, X, Y),
	(
		OldValue = SentinelValue,
		Influence = IncrFunc(X, Y, Influence0)
	->
		% Set the value for this neighbour, and insert it
		% into the frontier
		IM = set_square(IM0, X, Y, Influence),
		queue__put(Frontier0, frontier_elem(X, Y, Influence), Frontier)
	;
		% If this neighbour doesn't have the sentinel value, then
		% it has already been processed, so don't process it again.
		% Also, if the IncrFunc doesn't succeed, then it is not
		% reachable, so don't process it in that case either.
		IM = IM0,
		Frontier = Frontier0
	).

%-----------------------------------------------------------------------------%
