% The top-level

:- module main.
:- interface.
:- import_module io.

:- pred main(io__state::di, io__state::uo) is cc_multi.

%-----------------------------------------------------------------------------%

:- implementation.
:- import_module std_util, exception, list, bool.
:- import_module game, protocol, debug, strategy.
:- import_module globals, options, getopt, library.
:- import_module curs, output.

%-----------------------------------------------------------------------------%

% Handle command-line options

main -->
	io__command_line_arguments(Args0),
	{ OptionOps = option_ops(short_option, long_option, option_defaults,
				special_handler) },
	{ getopt__process_options(OptionOps, Args0, Args, Result0) },
	postprocess_options(Result0, Result),
	main_2(Result, Args).


:- pred postprocess_options(maybe_option_table(option),
	maybe(string), io__state, io__state).
:- mode postprocess_options(in, out, di, uo) is det.

postprocess_options(error(ErrorMessage), yes(ErrorMessage)) --> [].
postprocess_options(ok(OptionTable), no) --> 
	globals__io_init(OptionTable).

        % Display error message and then usage message
:- pred usage_error(string::in, io__state::di, io__state::uo) is det.
usage_error(ErrorMessage) -->
        io__progname_base("main", ProgName),
        io__stderr_stream(StdErr),
        io__write_string(StdErr, ProgName),
        io__write_string(StdErr, ": "),
        io__write_string(StdErr, ErrorMessage),
        io__write_string(StdErr, "\n"),
        io__set_exit_status(1),
        usage.

        % Display usage message
:- pred usage(io__state::di, io__state::uo) is det.
usage -->
        io__progname_base("main", ProgName),
        io__stderr_stream(StdErr),
	{ library__version(Version) },
        io__write_strings(StdErr,
		["Merry Mercurian's robot client, built with Mercury version ", Version, "\n"]),
        io__write_string(StdErr, "Copyright (C) 2002 The University of Melbourne\n"),
        io__write_string(StdErr, "Usage: "),
        io__write_string(StdErr, ProgName),
	io__write_string(StdErr, " [<options>] hostname port\n").
        %io__write_string(StdErr, "Use `"),
        %io__write_string(StdErr, ProgName),
        %io__write_string(StdErr, " --help' for more information.\n").

:- pred long_usage(io__state::di, io__state::uo) is det.
long_usage -->
	usage,
	io__write_string("\n"),
        io__write_string("Options:\n"),
        options_help.

:- pred main_2(maybe(string), list(string), io__state, io__state).
:- mode main_2(in, in, di, uo) is cc_multi.

main_2(yes(ErrorMessage), _) -->
        usage_error(ErrorMessage).
main_2(no, Args) -->
        globals__io_lookup_bool_option(help, Help),
        (
                { Help = yes }
        ->
                long_usage
        ;
		( { Args = [Host, Port] } ->
			connect_to_server(Host, Port)
		;
			io__write_string("Usage: main <hostname> <port>\n"),
			% by default, just run from stdin/stdout
			io__write_string("Entering debug mode - I/O via stdin/stdout\n")
		),

		% run it and ignore any exceptions
		% (it's normal to get an exception when the game ends)
		maybe_curses(curs__start),
		try_io(main_3, _ExceptionResult),
		maybe_curses(curs__stop)
	).

%-----------------------------------------------------------------------------%

% Do the real work

:- pred main_3(unit, io__state, io__state).
:- mode main_3(out, di, uo) is cc_multi.

main_3(unit) -->
	init(GameState0),
	run(GameState0).

init(Game) -->
	% The initialization phase occurs as soon as a client joins a game. The
	% protocol is as follows:
	%
	%  1.Client: send the string "Player"
	%  2.Server: send board
	%  3.Server: send player's configuration
	%  4.Server: send initial server response
	%    (with the initial location of all players)
	send_init,
	read_board(Board),
	read_config(Config),
	read_server_response(InitialResponse),
	{ Game = init_game(Board, Config, InitialResponse) }.
	%output_string("finished init\n").

% Note: this is an infinite loop.
% When the game ends, the server will close the connection,
% and we'll throw an exception.
:- mode run(in, di, uo) is cc_multi.
run(GameState0) -->
	output_state(GameState0), % for debugging
	%io__report_stats,
	do_turn(GameState0, GameState),
	run(GameState).

:- mode do_turn(in, out, di, uo) is cc_multi.
do_turn(GameState0, GameState) -->
	read_current_packages(Packages),
	{ update_packages(Packages, GameState0, GameState1) },
	{ choose_command(Packages, GameState1, Command, GameState2) },
	send_command(Command),
	maybe_curses(curs__refresh),
	read_server_response(Response),
	{ update_game_state(Response, GameState2, GameState) }.

:- pred choose_command(current_packages, game, command, game).
:- mode choose_command(in, in, out, out) is cc_multi.
choose_command(Packages, GameState0, Command, GameState) :-
	% first try the smart algorithm,
	% then if that fails, use the safe algorithm
	try(smart_algorithm(Packages, GameState0), ExceptionResult),
	(
		ExceptionResult = succeeded({Command, GameState})
	;
		ExceptionResult = exception(_),
		safe_algorithm(Packages, GameState0, Command),
		GameState = GameState0
	).

:- pred safe_algorithm(current_packages::in, game::in, command::out) is det.
safe_algorithm(_CurrentPackages, _GameState, Command) :-
	% XXX this is a really dumb algorithm...
	Command = nop_command.

%-----------------------------------------------------------------------------%

:- pred smart_algorithm(current_packages::in, game::in,
		{command,game}::out) is det.
smart_algorithm(CurrentPackages, GameState0, {Command, GameState}) :-
	strategy__choose_command(CurrentPackages, Command, GameState0,
		GameState).

%-----------------------------------------------------------------------------%
