%-----------------------------------------------------------------------------%
% This module implements the navigation strategy to be used by the robot.
%
% The strategy is based around an operator that takes a set of paths P
% from the current location and produces a new set of paths that can be
% reached from P by adding one extra step.
%
% For each path we collect certain information which can help us evaluate
% the path, e.g. the path length.  When two paths end at the same location,
% we prune the path which we evaluate to be the worst.
%
% If a step results in no new locations being reachable, then we have reached
% a fixpoint and we can select the best known path.  If we never remove any
% locations, then since there is a finite number of different locations on
% the board our algorithm must terminate.  However, we may wish to limit
% the number of iterations or otherwise prune the tree to improve performance.
%-----------------------------------------------------------------------------%

:- module navigate_2.
:- interface.
:- import_module game.
:- import_module list, set, std_util.

	% Values of this type represent paths starting from the current
	% position.  Note that the current position is implied; it is not
	% directly represented here.
	%
:- type path
	--->	path(
			path_info	:: path_info,
				% Accumulated information about the path.

			path_end	:: coord,
				% The final location.

			last_move	:: direction,
				% The final move before reaching the above
				% location.  Note carefully: we start the
				% iterations with a path that has the start
				% location as its path_end.  For this path,
				% the last_move field is meaningless.

			previous_moves	:: maybe(path)
				% If there were previous moves M then this
				% is yes(M), otherwise no.
		).

	% This type records the information that we accumulate for each
	% path.  We use this information to evaluate the path.  In this
	% version of our algorithm, the only information that we keep around
	% is the final cost of the path.  This restricts our choice of
	% cost function, but makes the algorithm a lot cheaper.
	%
:- type path_info == float.

	% A function of this type must be provided by the caller of the step
	% operator.  The reason this function is abstracted away is that
	% we may wish to use different strategies at different times.  E.g.:
	% 	- later in the game we may not consider the risk of water
	%	  to be as serious; 
	%	- when we hold valuable packages we may wish to avoid opponents
	%	  and open space.
	%
:- type value_func == (func(coord) = float).

%-----------------------------------------------------------------------------%

	% Values of this type represent sets of paths starting from the
	% current location.  This is the domain which our operator works on.
	%
:- type path_set
	--->	path_set(
			frontier	:: list(path),
				% New paths that have just been reached.

			reached		:: set(coord),
				% The set of all locations we have reached,
				% either in the current frontier or some time
				% earlier.

			best		:: maybe({path, float})
				% The best path seen so far, and its valuation.
				% If we haven't taken any steps yet, this
				% will be 'no'.
		).

%-----------------------------------------------------------------------------%

	% find_best_path(Game, ValueFunc, Goal, Limit) returns the best path
	% and value available, given the following:
	%	Game		game data
	%	ValueFunc	evaluates paths.
	%	Goal		a specific destination
	%	Limit		maximum steps allowed.
	%
	% If there are no paths within the specified limit, return 'no'.
	%
:- func find_best_path(game, value_func, coord, int) = maybe(path).

	% Returns the list of moves in the order they should be performed.
	%
:- func get_moves(path) = list(direction).

%-----------------------------------------------------------------------------%

:- implementation.
:- import_module int, float, exception.

find_best_path(Game, ValueFunc, Goal, Limit) = Best :-
	PathSet0 = init_path_set(Game ^ my_robot),
	Best = iterate_step_to_fixpoint(Game, ValueFunc, Goal, Limit, PathSet0).

get_moves(Path) = get_moves_2(Path, []).

:- func get_moves_2(path, list(direction)) = list(direction).

get_moves_2(Path, Moves0) = Moves :-
	MaybePreviousMoves = Path ^ previous_moves,
	(
		MaybePreviousMoves = yes(PreviousMoves),
		Moves = get_moves_2(PreviousMoves, [Path ^ last_move | Moves0])
	;
		MaybePreviousMoves = no,
		% Don't insert the last_move of this path, since it is a dummy.
		Moves = Moves0
	).

%-----------------------------------------------------------------------------%

:- func init_path_set(my_robot_info) = path_set.

init_path_set(MyInfo) = PathSet :-
	Pos = MyInfo ^ my_posn,
	PathInfo = init_path_info(MyInfo),
	Dummy = north,
	Frontier = [path(PathInfo, Pos, Dummy, no)],
	PathSet = path_set(Frontier, set__init, no).

:- func init_path_info(my_robot_info) = path_info.

init_path_info(_MyInfo) = 0.0.

%-----------------------------------------------------------------------------%

:- func iterate_step_to_fixpoint(game, value_func, coord, int, path_set)
	= maybe(path).

iterate_step_to_fixpoint(Game, ValueFunc, Goal, Count, PathSet0) = Best :-
	(
		Count > 0
	->
		Best = iterate_to_fixpoint_2(Game, ValueFunc, Goal, Count,
			PathSet0)
	;
		Best = no
	).

:- func iterate_to_fixpoint_2(game, value_func, coord, int, path_set)
	= maybe(path).

iterate_to_fixpoint_2(Game, ValueFunc, Goal, Count, PathSet0) = MaybeBest :-
	Frontier0 = PathSet0 ^ frontier,
	(
		Frontier0 = [],
		%
		% Reached a fixpoint, so return.
		%
		MaybeBest = no
	;
		Frontier0 = [_ | _],
		%
		% Calculate the new frontier.
		%
		Reached0 = PathSet0 ^ reached,
		calculate_new_frontier(Game, ValueFunc, Goal, Reached0,
			MaybeFound, Frontier0, Frontier),

		(
			MaybeFound = no,
			%
			% Now determine the other fields of the path set.
			%
			Reached = add_locations(Reached0, Frontier),
			Best0 = find_best_path(PathSet0 ^ best, Frontier),

			%
			% Keep iterating with the new path set.
			%
			PathSet1 = ((PathSet0 ^ best := Best0)
					^ frontier := Frontier)
					^ reached := Reached,
			MaybeBest = iterate_step_to_fixpoint(Game, ValueFunc,
				Goal, Count - 1, PathSet1)
		;
			MaybeFound = yes(_),
			MaybeBest = MaybeFound
		)
	).

%-----------------------------------------------------------------------------%

% When calculating the new frontier, we do the processing in two stages.
%
% The first stage is to calculate all of the new locations, and associate
% with each the path and direction that was used to get there.  Then we
% remove duplicate locations, and locations we have already been to.
%
% The second stage is to calculate new paths that lead to these new
% locations.  The paths have extra information, the path_infos, that we
% need to calculate for each new location.
%
% The reason we need to split this process into two stages is that we want
% to remove duplicates as soon as possible, before we calculate the
% path_infos.  If we don't do this, we will calculate a lot of path_infos
% unnecessarily.
%
% The intermediate_frontier type is used to hold the intermediate data
% produced by the first stage.  The second stage processes this data to
% produce a list of paths.

:- type intermediate_frontier == list({coord, direction, path}).

:- pred calculate_new_frontier(game::in, value_func::in, coord::in,
	set(coord)::in, maybe(path)::out, list(path)::in,
	list(path)::out) is det.

calculate_new_frontier(Game, ValueFunc, Goal, Reached, MaybeFound, Frontier0,
		Frontier) :-
	%
	% stage 1
	%
	Intermediate0 = do_one_step(Game ^ board, Frontier0),
	list__sort(compare_first, Intermediate0, Intermediate1),
	remove_adjacent_equivs(Intermediate1, Intermediate2),
	Intermediate = remove_reached_locations(Reached, Intermediate2),
	%
	% stage 2
	%
	process_intermediate_frontier(Game, ValueFunc, Goal, Intermediate, [],
		RevFrontier, MaybeFound),
	Frontier = list__reverse(RevFrontier).

	% From each path in the given set take one step in each direction
	% possible.  If the resulting location is not on a wall or on water,
	% add the tuple of information to the output list.  We look for the
	% goal here.  If found, we return straight away.
	%
	% The output list is unsorted.
	%
:- func do_one_step(board, list(path)) = intermediate_frontier.

do_one_step(Board, Paths) = list__foldl(do_one_step_path(Board), Paths, []).

:- func do_one_step_path(board, path, intermediate_frontier)
	= intermediate_frontier.

do_one_step_path(Board, Path, Ps0) = Ps :-
	Ps1 = do_one_step_direction(Board, north, Path, Ps0),
	Ps2 = do_one_step_direction(Board, south, Path, Ps1),
	Ps3 = do_one_step_direction(Board, east, Path, Ps2),
	Ps = do_one_step_direction(Board, west, Path, Ps3).

:- func do_one_step_direction(board, direction, path, intermediate_frontier)
	= intermediate_frontier.

do_one_step_direction(Board, Direction, Path, Ps0) = Ps :-
	Pos = move(Path ^ path_end, Direction),
	(
		square_is_free(Board, Pos)
	->
		Ps = [{Pos, Direction, Path} | Ps0]
	;
		Ps = Ps0
	).

	% square_is_free(Board, Pos) is true iff Pos is on the board and
	% the square kind there is plain or home.
:- pred square_is_free(board, coord).
:- mode square_is_free(in, in) is semidet.

square_is_free(Board, Pos) :-
	Kind = safe_square(Board, Pos),
	(
		Kind = plain
	;
		Kind = home
	).

:- pred compare_first({A, B, C}, {A, B, C}, comparison_result).
:- mode compare_first(in, in, out) is det.

compare_first({A, _, _}, {B, _, _}, Result) :-
	compare(Result, A, B).

	% Removes equivalent elements of the list which are adjacent
	% to one another.  We remove the most expensive path and keep
	% the cheapest one.
:- pred remove_adjacent_equivs(intermediate_frontier, intermediate_frontier).
:- mode remove_adjacent_equivs(in, out) is det.

remove_adjacent_equivs([], []).
remove_adjacent_equivs([X | Xs], L) :-
	remove_adjacent_equivs_2(Xs, X, L).

:- pred remove_adjacent_equivs_2(intermediate_frontier,
	{coord, direction, path}, intermediate_frontier).
:- mode remove_adjacent_equivs_2(in, in, out) is det.

remove_adjacent_equivs_2([], X, [X]).
remove_adjacent_equivs_2([X1 | Xs], X0, L) :-
	X0 = {Coord0, _Dir0, Path0},
	X1 = {Coord1, _Dir1, Path1},
	(
		Coord0 = Coord1
	->
		(
			Path0 ^ path_info > Path1 ^ path_info
		->
			remove_adjacent_equivs_2(Xs, X1, L)
		;
			remove_adjacent_equivs_2(Xs, X0, L)
		)
	;
		L = [X0 | L0],
		remove_adjacent_equivs_2(Xs, X1, L0)
	).

	% remove_reached_locations(Reached, Paths) is the list of paths
	% which are members of Paths and do not have their first element
	% in Reached.  The list order must be preserved.
:- func remove_reached_locations(set(coord), intermediate_frontier)
	= intermediate_frontier.

remove_reached_locations(Reached, Ps) = list__filter(NotReached, Ps) :-
	NotReached = (pred({A, _, _}::in) is semidet :-
			\+ set__member(A, Reached)
		).

	% Calculate path_infos for each member of the intermediate_frontier,
	% and build the new paths.  The output is in reverse order.  Check
	% for the goal as we go.
:- pred process_intermediate_frontier(game::in, value_func::in, coord::in,
	intermediate_frontier::in, list(path)::in, list(path)::out,
	maybe(path)::out) is det.

process_intermediate_frontier(_, _, _, [], Frontier, Frontier, no).
process_intermediate_frontier(Game, ValueFunc, Goal, Intermediate, Frontier0,
	Frontier, MaybeFound) :-

	Intermediate = [{Pos, Dir, Path0} | Intermediate0],
	PathInfo = ValueFunc(Pos) + (Path0 ^ path_info),
	Path = path(PathInfo, Pos, Dir, yes(Path0)),
	(
		Pos = Goal
	->
		MaybeFound = yes(Path),
		Frontier = []
	;
		process_intermediate_frontier(Game, ValueFunc, Goal,
			Intermediate0, [Path | Frontier0], Frontier, MaybeFound)
	).

%-----------------------------------------------------------------------------%

	% Add the ends of the new paths to the set of reached locations.
:- func add_locations(set(coord), list(path)) = set(coord).

add_locations(Reached0, Frontier) = Reached :-
	Reached1 = set__sorted_list_to_set(list__map(get_path_end, Frontier)),
	Reached = set__union(Reached0, Reached1).

:- func get_path_end(path) = coord.

get_path_end(P) = P ^ path_end.

:- func find_best_path(maybe({path, float}), list(path)) = maybe({path, float}).

find_best_path(no, []) = no.
find_best_path(no, [P | Ps]) = yes(BestVal) :-
	Value = P ^ path_info,
	BestVal = find_best_path_2(P, Value, Ps).
find_best_path(yes({Best0, Val0}), Ps) = yes(BestVal) :-
	BestVal = find_best_path_2(Best0, Val0, Ps).

:- func find_best_path_2(path, float, list(path)) = {path, float}.

find_best_path_2(Best, Val, []) = {Best, Val}.
find_best_path_2(Best0, Val0, [P | Ps]) = BestVal :-
	ValP = P ^ path_info,
	(
		ValP > Val0
	->
		BestVal = find_best_path_2(P, ValP, Ps)
	;
		BestVal = find_best_path_2(Best0, Val0, Ps)
	).

%-----------------------------------------------------------------------------%

