%-----------------------------------------------------------------------------%
% The purpose of this module is to find the best path to a goal, in the
% presence of potential danger.  The amount of danger at a particular
% square is provided by the caller in the form of a cost function.
% For the single robot case there is no danger, so the best path is simply
% the shortest path; hence this module is not required.  When other robots
% are present, however, it may be risky to go near them, or to go near other
% things (e.g. water).
%
% This module uses a "best first" strategy.
%-----------------------------------------------------------------------------%

:- module path.
:- interface.
:- import_module list.
:- import_module game.

:- type search_result
	--->	found(float, list(direction))
			% We found the goal at the end of the given path,
			% with the given cost.  The directions are listed in
			% the order that they should be taken.

	;	search_aborted.
			% The goal is not reachable by a path shorter
			% than the threshold provided.

	% This is how the costs of squares are determined.  The cost of a
	% path is equal to the cost of the squares covered by it, not
	% counting the first square.
	%
:- type square_cost_func == (func(coord) = float).

	% find_best_path(Board, Start, Goal, CostFunc, Cutoff) tries to find
	% the cheapest path from Start to Goal, following a valid path on
	% Board, with square cost function CostFunc.  Only paths whose cost
	% is lower than Cutoff are considered.  The result of the search
	% is returned.
	%
:- func find_best_path(board, coord, coord, square_cost_func, float)
	= search_result.

%-----------------------------------------------------------------------------%

:- implementation.
:- import_module bool, float, pqueue, tree234, std_util.

	% The frontier is the set of paths which may have neighbours
	% that haven't been searched yet.  We store it as a priority
	% queue, with the path costs used as the keys.  Each path is
	% stored as the list of directions in reverse order.
	%
:- type frontier == pqueue(float, path).

:- type path
	--->	path(
			path_end	:: coord,
				% End of the path.

			reverse_moves	:: list(direction)
				% Directions to go from the start of the
				% path to the finish.  The list is in
				% reverse order.
		).

:- type search_info
	--->	search_info(
			board		:: board,
			goal		:: coord,
			cost_func	:: square_cost_func,
			threshold	:: float
		).

%-----------------------------------------------------------------------------%

find_best_path(Board, Start, Goal, CostFunc, Cutoff) = Result :-
	SearchInfo = search_info(Board, Goal, CostFunc, Cutoff),
	tree234__set(tree234__init, Start, unit, Reached),
	Path = path(Start, []),
	Frontier = pqueue__insert(pqueue__init, 0.0, Path),
	Result = find_best_path_iterate(SearchInfo, Reached, Frontier).

:- type reached_set == tree234(coord, unit).

:- func find_best_path_iterate(search_info, reached_set, frontier)
	= search_result.

find_best_path_iterate(Info, Reached0, Frontier0) = Result :-
	(
		pqueue__remove(Frontier0, Cost, Path, Frontier1)
	->
		search_neighbours(Info, Reached0, Frontier1, Cost, Path,
			[north, south, east, west], Reached2, Frontier2, Found),
		(
			Found = yes({GoalCost, GoalPath}),
			Moves = list__reverse(GoalPath ^ reverse_moves),
			Result = found(GoalCost, Moves)
		;
			Found = no,
			Result = find_best_path_iterate(Info, Reached2,
				Frontier2)
		)
	;
		Result = search_aborted
	).

:- pred search_neighbours(search_info::in, reached_set::in, frontier::in,
	float::in, path::in, list(direction)::in, reached_set::out,
	frontier::out, maybe({float, path})::out) is det.

search_neighbours(_, Reached, Frontier, _, _, [], Reached, Frontier, no).
search_neighbours(Info, Reached0, Frontier0, Cost, Path, [Dir | Dirs],
	Reached, Frontier, Found) :-

	search_neighbour(Info, Reached0, Frontier0, Cost, Path, Dir, Reached1,
		Frontier1, Found0),
	(
		Found0 = yes(_),
		Reached = Reached1,
		Frontier = Frontier1,
		Found = Found0
	;
		Found0 = no,
		search_neighbours(Info, Reached1, Frontier1, Cost, Path, Dirs,
			Reached, Frontier, Found)
	).

:- pred search_neighbour(search_info::in, reached_set::in, frontier::in,
	float::in, path::in, direction::in, reached_set::out, frontier::out,
	maybe({float, path})::out) is det.

search_neighbour(Info, Reached0, Frontier0, Cost0, Path0, Dir, Reached,
	Frontier, Found) :-

	Location = move(Path0 ^ path_end, Dir),
	(
		square_is_open(Info ^ board, Location),
		\+ tree234__search(Reached0, Location, _)
	->
		Cost = Cost0 + (Info ^ cost_func)(Location),
		Path = path(Location, [Dir | Path0 ^ reverse_moves]),
		(
			Location = Info ^ goal
		->
			Reached = Reached0,
			Frontier = Frontier0,
			Found = yes({Cost, Path})
		;
			Cost < Info ^ threshold
		->
			Reached = tree234__set(Reached0, Location, unit),
			pqueue__insert(Frontier0, Cost, Path, Frontier),
			Found = no
		;
			Reached = Reached0,
			Frontier = Frontier0,
			Found = no
		)
	;
		Reached = Reached0,
		Frontier = Frontier0,
		Found = no
	).

:- pred square_is_open(board::in, coord::in) is semidet.

square_is_open(Board, Coord) :-
	Kind = square(Board, Coord),
	(
		Kind = plain
	;
		Kind = home
	).

