%-----------------------------------------------------------------------------%
:- module protocol.
:- interface.
:- import_module io, game.


:- type hostname == string.
:- type port == string.
:- pred connect_to_server(hostname::in, port::in, io__state::di, io__state::uo)
	is det.

% The initialization phase occurs as soon as a client joins a game. The
% protocol is as follows:
%
%  1.Client: send the string "Player"
%  2.Server: send board 
%  3.Server: send player's configuration 
%  4.Server: send initial server response
%    (with the initial location of all players) 

:- pred send_init(io__state::di, io__state::uo) is det.

:- pred read_board(board::out, io__state::di, io__state::uo) is det.

:- pred read_config(config::out, io__state::di, io__state::uo) is det.

:- pred read_server_response(server_response::out, io__state::di, io__state::uo)
	is det.

% The game is made up of a number of turns. A turn looks as follows: 
% 
%   1.Server: a list of packages available at the player's current location 
%   2.Client: issue a command to the robot 
%   3.Server: server responds with what happened during the turn 

:- pred read_current_packages(current_packages::out,
	io__state::di, io__state::uo) is det.

:- pred send_command(command::in, io__state::di, io__state::uo) is det.

% (read_server_response already declared above)

%-----------------------------------------------------------------------------%
:- implementation.
:- import_module int, list, exception, char, string, curs.
:- import_module output.

%-----------------------------------------------------------------------------%

send_init -->
	io__write_string("Player\n"),
	io__flush_output.

%-----------------------------------------------------------------------------%

:- pred read_line_from_socket(io__result(list(char)), io__state, io__state).
read_line_from_socket(Result) -->
	% print to stderr, for debugging...
	%io__stderr_stream(Stderr),
	%io__write_string(Stderr, "reading line...\n"),

	io__read_line(Result),

	%io__print(Stderr, "read line: ["),
	%io__print(Stderr, Result),
	%io__print(Stderr, "]."),
	%io__nl(Stderr),
	[].

%-----------------------------------------------------------------------------%

read_board(Board) -->
	read_board_dimensions(X, Y),
	read_rows(X, Y, Rows),
	{ Board = make_board(X, Y, Rows) }.

read_board_dimensions(X, Y) -->
	read_line_from_socket(Result),
	( { Result = ok(Line) } ->
		{ parse_board_dimensions(Line, (X,Y)) }
	;
		{ throw(Result) }
	).

parse_board_dimensions(Line, Dim) :-
	( p_board_dimensions(Dim0, Line, []) ->
		Dim = Dim0
	;
		throw("board dimensions parse error")
	).

p_board_dimensions((x(X),y(Y))) -->
	p_int(X),
	p_space,
	p_int(Y),
	p_nl.

read_rows(X, y(Y), Rows) -->
	( { Y = 0 } ->
		{ Rows = [] }
	;
		read_row(X, Row),
		read_rows(X, y(Y - 1), Rest),
		{ Rows = [Row | Rest] }
	).

read_row(X, Row) -->
	read_line_from_socket(Result),
	( { Result = ok(Line) } ->
		{ parse_row(X, Line, Row) }
	;
		{ throw(Result) }
	).

parse_row(X, Line, Row) :-
	( p_row(X, Row0, Line, []) ->
		Row = Row0
	;
		throw("row parse error")
	).

p_row(x(X), Row) -->
	( { X = 0 } ->
		p_nl,
		{ Row = [] }
	;
		p_square(Square),
		p_row(x(X - 1), Rest),
		{ Row = [Square | Rest] }
	).

p_square(Square) -->
	[Char],
	( { Char = ('.') },
		{ Square = plain }
	; { Char = ('~') },
		{ Square = lethal }
	; { Char = ('#') },
		{ Square = wall }
	; { Char = ('@') },
		{ Square = home }
	).

%-----------------------------------------------------------------------------%

read_config(Config) -->
	read_line_from_socket(Result),
	( { Result = ok(Line) } ->
		{ parse_config(Line, Config) }
	;
		{ throw(Result) }
	).

parse_config(Line, Config) :-
	( p_config(Config0, Line, []) ->
		Config = Config0
	;
		throw("config parse error")
	).

p_config(config(robot_id(MyId), MyCapacity, MyMoney)) -->
	p_int(MyId),
	p_space,
	p_int(MyCapacity),
	p_space,
	p_int(MyMoney),
	p_nl.

%-----------------------------------------------------------------------------%

read_server_response(Response) -->
	read_line_from_socket(Result),
	( { Result = ok(Line) } ->
		parse_response(Line, Response)
	;
		{ throw(Result) }
	).

parse_response(Line, Response) -->
	( { p_update_list_nl(Updates, Line, []) } ->
		{ Response = Updates }
	;
		handle_unexpected_response(Line),
		% { throw("response parse error") }
		{ Response = [] }
	).

p_update_list_nl(Updates) --> p_update_list(Updates), p_nl.

p_update_list(Updates) -->
	( p_update(Update) ->
		( p_space, p_update_list(Rest), { Rest \= [] } ->
			{ Updates = [Update|Rest] }
		;
			{ Updates = [Update] }
		)
	;
		{ Updates = [] }
	).

p_update(robot_update(RobotId, Actions)) -->
	p_robot_id(RobotId),
	p_space,
	p_action_list(Actions).

p_robot_id(robot_id(Id)) -->
	['#'],
	p_int(Id).

p_action_list(Actions) -->
	( p_action(Action) ->
		( p_space, p_action_list(Rest), { Rest \= [] } ->
			{ Actions = [Action|Rest] }
		;
			{ Actions = [Action] }
		)
	;
		{ Actions = [] }
	).

:- pred p_action(robot_action, list(char), list(char)).
p_action(Action) -->
	[Char],
	( { Char = 'N' },
		{ Action = moved(north) }
	; { Char = 'S' },
		{ Action = moved(south) }
	; { Char = 'E' },
		{ Action = moved(east) }
	; { Char = 'W' },
		{ Action = moved(west) }
	; { Char = 'P' },
		p_space,
		p_int(PId),
		{ Action = picked(package_id(PId)) }
	; { Char = 'D' },
		p_space,
		p_int(PId),
		{ Action = dropped(package_id(PId)) }
	; { Char = 'X' },
		p_space,
		p_int(X),
		p_space,
		['Y'],
		p_space,
		p_int(Y),
		{ Action = position(x(X),y(Y)) }
	).

%-----------------------------------------------------------------------------%

read_current_packages(Packages) -->
	read_line_from_socket(Result),
	( { Result = ok(Line) } ->
		parse_packages(Line, Packages)
	;
		{ throw(Result) }
	).

parse_packages(Line, Packages) -->
	( { p_package_list_nl(Packages0, Line, []) } ->
		{ Packages = Packages0 }
	;
		handle_unexpected_response(Line),
		% { throw("package parse error") }
		{ Packages = [] }
	).

handle_unexpected_response(Line) -->
	% This may be a message that the robot died.
	% Just copy it to stderr.
	maybe_curses(curs__stop),
	io__stderr_stream(Stderr),
	io__set_output_stream(Stderr, Old),
	io__write_string("Response from server:\n"),
	io__write_string(string__from_char_list(Line)),
	io__flush_output,
	cat,
	io__set_output_stream(Old, _).

:- pred cat(io__state::di, io__state::uo) is det.

cat -->
	io__read_line_as_string(Result),
	( { Result = ok(Line) },
		io__write_string(Line),
		io__flush_output,
		cat
	; { Result = eof }
	; { Result = error(_) },
		{ throw(Result) }
	).

%-----------------------------------------------------------------------------%

p_package_list_nl(Packages) --> p_package_list(Packages), p_nl.

p_package_list(Packages) -->
	( p_package(Package) ->
		( p_space, p_package_list(Rest), { Rest \= [] } ->
			{ Packages = [Package|Rest] }
		;
			{ Packages = [Package] }
		)
	;
		{ Packages = [] }
	).

p_package(current_package(package_id(Id), x(X), y(Y), Weight)) -->
	p_int(Id),
	p_space,
	p_int(X),
	p_space,
	p_int(Y),
	p_space,
	p_int(Weight).

p_int(Int) -->
	p_digits(Digits), { Digits \= [] },
	{ Int = string__det_to_int(string__from_char_list(Digits)) }.

% a list of digits, possibly empty
p_digits(Digits) -->
	( [Digit], { char__is_digit(Digit) } ->
		p_digits(Rest),
		{ Digits = [Digit|Rest] }
	;
		{ Digits = [] }
	).

% zero or more spaces
p_space --> ( [' '] -> p_space ; [] ).

% a newline, optionally preceded by spaces
p_nl --> ['\n'] ; [' '], p_nl.

%-----------------------------------------------------------------------------%

send_command(Command) -->
	dump_command(Command),
	send_command_2(Command),
	io__flush_output.

:- pred send_command_2(command::in, io__state::di, io__state::uo) is det.
send_command_2(command(Bid, Action)) -->
	io__write_int(Bid),
	io__write_char(' '),
	send_action(Action),
	io__write_char('\n').

:- pred send_action(action::in, io__state::di, io__state::uo) is det.
send_action(move(Direction)) -->
	io__write_string("Move "),
	send_direction(Direction).
send_action(pick(PackageIds)) -->
	io__write_string("Pick "),
	send_package_list(PackageIds).
send_action(drop(PackageIds)) -->
	io__write_string("Drop "),
	send_package_list(PackageIds).

:- pred send_direction(direction::in, io__state::di, io__state::uo) is det.
send_direction(north) --> io__write_string("N").
send_direction(south) --> io__write_string("S").
send_direction(east) --> io__write_string("E").
send_direction(west) --> io__write_string("W").

:- pred send_package_list(list(package_id)::in, io__state::di, io__state::uo)
	is det.
send_package_list(Packages) -->
	io__write_list(Packages, " ", send_package).

:- pred send_package(package_id::in, io__state::di, io__state::uo) is det.
send_package(package_id(Id)) -->
	io__write_int(Id).

dump_command(command(Bid, Action)) -->
	output_string(string__int_to_string(Bid)),
	output_string(" "),
	dump_action(Action),
	output__nl.

:- pred dump_action(action::in, io__state::di, io__state::uo) is det.
dump_action(move(Direction)) -->
	output_string("Move "),
	dump_direction(Direction).
dump_action(pick(PackageIds)) -->
	output_string("Pick "),
	dump_package_list(PackageIds).
dump_action(drop(PackageIds)) -->
	output_string("Drop "),
	dump_package_list(PackageIds).

:- pred dump_direction(direction::in, io__state::di, io__state::uo) is det.
dump_direction(north) --> output_string("N").
dump_direction(south) --> output_string("S").
dump_direction(east) --> output_string("E").
dump_direction(west) --> output_string("W").

:- pred dump_package_list(list(package_id)::in, io__state::di, io__state::uo)
	is det.
dump_package_list(Packages) -->
	io__write_list(Packages, "", dump_package).

:- pred dump_package(package_id::in, io__state::di, io__state::uo) is det.
dump_package(package_id(Id)) -->
	output_string(string__int_to_string(Id)),
	output_string(" ").

%-----------------------------------------------------------------------------%

:- pragma c_code(connect_to_server(HostName::in, Port::in, _IO0::di, _IO::uo),
	[will_not_call_mercury, promise_pure],
"
	connect_to_socket(HostName, Port);
").

:- pragma c_header_code("
/* Connect to the specified socket, and set Mercury's stdin/stdout
   to point to it. */
void connect_to_socket(const char *hostname, const char *port_string);

void fatal_error(const char *msg);

#include ""mercury_library_types.h""

extern MercuryFile mercury_stdin;
extern MercuryFile mercury_stdout;

#include <stdio.h>
#include <errno.h>
#include <stdarg.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <netdb.h>
#include <stdlib.h>                                                             

").

:- pragma c_code("
int debug = 0;

void fatal_error(const char *msg) {
	fprintf(stderr, msg);
	exit(1);
}

void
connect_to_socket(const char *hostname, const char *port_string)
{
	int fd;
	int len;
	FILE *file_in;
	FILE *file_out;
	int addr_family;
	char *inet_socket;
	struct sockaddr_in inet_address;
	struct sockaddr* addr;

	unsigned short port;
	int host_addr;

	/*
	** Set up the socket address.
	*/

	host_addr = inet_addr(hostname);
	if (host_addr == -1) {
		fatal_error(""connect_to_socket: invalid address"");
	}
	if (sscanf(port_string, ""%hu"", &port) != 1) {
		fatal_error(""connect_to_socket: invalid port"");
	}

	if (debug) {
		fprintf(stderr, ""connect_to_socket: host = %s, port = %d\n"",
			hostname, port);
	}
	inet_address.sin_family = AF_INET;
	inet_address.sin_addr.s_addr = host_addr;
	inet_address.sin_port = htons(port);
	addr_family = AF_INET;
	addr = (struct sockaddr *) &inet_address;
	len = sizeof(inet_address);

	/*
	** Open the socket.
	*/

	fd = socket(addr_family, SOCK_STREAM, 0);
	if (fd < 0) {
		fprintf(stderr, ""connect_to_socket: socket() failed: %s\n"",
			strerror(errno));
		fatal_error(""cannot open socket"");
	} else if (debug) {
		fprintf(stderr,""connect_to_socket: creation of socket ok\n"");
	}

	/*
	** Connect to the socket
	*/
	if (connect(fd, addr, len) < 0) {
		fprintf(stderr, ""connect_to_socket: connect() failed: %s\n"",
			strerror(errno));
		fatal_error(""can't connect to socket"");
	} else if (debug) {
		fprintf(stderr, ""connect_to_socket: connection to socket: ok\n"");
	}

	/*
	** Convert the socket fd to a Mercury stream
	*/
	file_in = fdopen(fd, ""r"");
	file_out = fdopen(fd, ""w"");
	if ((file_in == NULL)||(file_out == NULL)) {
		fprintf(stderr, ""connect_to_socket: fdopen() failed: %s\n"",
			strerror(errno));
		fatal_error(""cannot open socket"");
	} else if (debug) {
		fprintf(stderr, ""connect_to_socket: fdopen(): ok\n"");
	}

	MR_mercuryfile_init(file_in, 1, &mercury_stdin);
	MR_mercuryfile_init(file_out, 1, &mercury_stdout);

}
").
